#!/usr/bin/env python
# -*- coding: iso-8859-15 -*-h

############################################################
#
#      (c)   Jean-Matthieu Haussaire, Marc Bocquet
#      Programme l95grs_release.py
#      2014-2016
#
#      Supplementary material for:
#      J.-M. Haussaire and M. Bocquet
#      "A low-order coupled chemistry meteorology model
#      for testing online and offline data assimilation schemes:
#      L95-GRS (v1.0)"
#      published in Geophysical Model Development
#      doi:10.5194/gmd-9-393-2016
#  
#      Last change: 21/02/2015
#
############################################################


import numpy as np
import random
from math import sqrt, exp


# Timer
time = 0

# Dimensions
G = 40 # Grid size
Nsp = 5 # Number of species
M = G*(1+Nsp) # Full state space size

# Lorenz 95 constants
F = 8.

# Rosenbrock constant
gamma = 1.+1./sqrt(2.);

# Chemical constants
k2 = 12.3    # at 300K, in ppb^{-1}.min^{-1}
k4 = 0.0275 # at 300K, in ppb^{-1}.min^{-1}
k5 = 10.2      # in ppb^{-1}.min^{-1}
k6 = 2*0.12    # in ppb^{-1}.min^{-1}
# Ozone photolysis for each hour of the day (over 24)
# for the 21st of March, in min^{-1}
k3t = [0.000112588,           
       0.00328719,
       0.00651789,
       0.00845721,
       0.00959167,
       0.0101921,
       0.0103804,
       0.0103804,
       0.0101921,
       0.00959167,
       0.00845721,
       0.00651789,
       0.00328719,
       0.000112588,
       0,0,0,0,0,0,0,0,0,0]
k3t = 60*np.array(k3t)

# Conversion factor from day^{-1} to Lorenz95 units^{-1}
conv = 5

# Deposition velocity in day^{-1} -> Lorenz95 units^{-1}
vdep = 0.02*conv
vdep_ROC = vdep
vdep_NO = vdep
vdep_NO2 = vdep
vdep_O3 = vdep
vdep_SGN = vdep

# Emissions in ppb.day^{-1} -> ppb.Lorenz95 units^{-1}
E_NOx = 0.27*conv
E_NO = E_NOx*0.9
E_NO2 = E_NOx*0.1
E_ROC = 0.0235*conv
# E_ROC = 0.00235*conv

# NO2 photolysis rate
def k3_def(time):

    # time: absolute time in Lorenz95 units (1 = 120h)
    # index in the k3 table
    index = int(time*120)%24

    return k3t[index]  # in min^{-1}
     
# ROC photolysis rate
def k1_def(k3):

    T = 300.
    k1 = 10000*exp(-4710./T)*k3

    return k1  # in min^{-1}
 

####################
# Evolution Model  #
####################

# Integration time-step
# 1 hour
dt = 0.05/6.

# If the Rosenbrock scheme adaptive or not
adaptive_option = True


# Rate for the non-stiff part of the model
# to be integrated by the RK4 scheme
def KineticNonStiff(x):

    # The time unit is Lorenz95'

    x_dot = np.zeros([M])

    wind = x[0:G]
    ROC  = x[G:2*G]
    NO   = x[2*G:3*G]
    NO2  = x[3*G:4*G]
    O3   = x[4*G:5*G]
    SGN  = x[5*G:]


    # Compute the fluxes
    phi_ROC = np.empty(G)
    phi_NO = np.empty(G)
    phi_NO2 = np.empty(G)
    phi_O3 = np.empty(G)
    phi_SGN = np.empty(G)
   
    for g in range(G):
        if wind[g] >= 0:
            phi_ROC[g] = wind[g]*ROC[g]
            phi_NO[g] = wind[g]*NO[g]
            phi_NO2[g] = wind[g]*NO2[g]
            phi_O3[g] = wind[g]*O3[g]
            phi_SGN[g] = wind[g]*SGN[g]
        else :
            phi_ROC[g] = wind[g]*ROC[(g+1)%G]
            phi_NO[g] = wind[g]*NO[(g+1)%G]
            phi_NO2[g] = wind[g]*NO2[(g+1)%G]
            phi_O3[g] = wind[g]*O3[(g+1)%G]
            phi_SGN[g] = wind[g]*SGN[(g+1)%G]
             

    for g in range(G):
        # wind
        x_dot[g] = (wind[(g+1)%G]-wind[(g-2+G)%G])*wind[(g-1+G)%G]-wind[g]+F;
        # ROC
        x_dot[G+g] = phi_ROC[(g-1+G)%G] - phi_ROC[g] - vdep_ROC*ROC[g] + E_ROC
        # NO
        x_dot[2*G+g] = phi_NO[(g-1+G)%G] - phi_NO[g] - vdep_NO*NO[g] + E_NO
        # NO2
        x_dot[3*G+g] = phi_NO2[(g-1+G)%G] - phi_NO2[g] - vdep_NO2*NO2[g] + E_NO2
        # O3
        x_dot[4*G+g] = phi_O3[(g-1+G)%G] - phi_O3[g] - vdep_O3*O3[g]
        # SGN
        x_dot[5*G+g] = phi_SGN[(g-1+G)%G] - phi_SGN[g] - vdep_SGN*SGN[g]

    return x_dot


# Rate for the stiff part of the model
# to be integrated by the Rosenbrock scheme
def KineticStiff(Z):

    # The time unit is min

    x_dot = np.zeros([G*Nsp])

    k3 = k3_def(time)
    k1 = k1_def(k3)

    ROC  = Z[:G]
    RP   = np.empty(G)
    NO   = Z[G:2*G]
    NO2  = Z[2*G:3*G]
    O3   = Z[3*G:4*G]
    SGN  = Z[4*G:]

    for g in range(G):
        # ROC
        x_dot[g] = 0.
        # RP
        b = (k2*NO[g]+k6*NO2[g])/k5
        c = -k1/k5*ROC[g]
        RP[g] = (-b+sqrt(b*b-4*c))/2.
        # NO
        x_dot[G+g] = NO[g]*(-k2*RP[g]-k4*O3[g])+k3*NO2[g]
        # NO2
        x_dot[2*G+g] = NO2[g]*(-k3-k6*RP[g])+k4*NO[g]*O3[g]+k2*RP[g]*NO[g]
        # O3
        x_dot[3*G+g] = k3*NO2[g]-k4*NO[g]*O3[g]
        # SGN
        x_dot[4*G+g] = k6*RP[g]*NO2[g]

    return x_dot


# Jacobian of the stiff part of the model
# for use in the Rosenbrock integration
def Jacobian(Z):

    jac = np.zeros([G*Nsp,G*Nsp])

    k3 = k3_def(time)
    k1 = k1_def(k3)

    ROC  = Z[:G]
    RP   = np.empty(G)
    NO   = Z[G:2*G]
    NO2  = Z[2*G:3*G]
    O3   = Z[3*G:4*G]
    SGN  = Z[4*G:]

    for g in range(G):

        # Diagnostic of RP
        b = (k2*NO[g]+k6*NO2[g])/k5
        c = -k1/k5*ROC[g]
        RP[g] = (-b+sqrt(b*b-4*c))/2.
    
        # df/dNO
        jac[G+g,G+g] = -k4*O3[g]-k2*RP[g]
        jac[2*G+g,G+g] = k2*RP[g]+k4*O3[g]
        jac[3*G+g,G+g] = -k4*O3[g]
        
        # df/dNO2
        jac[G+g,2*G+g] = k3
        jac[2*G+g,2*G+g] = -k3-k6*RP[g]
        jac[3*G+g,2*G+g] = k3
        jac[4*G+g,2*G+g] = k6*RP[g]

        # df/dO3
        jac[G+g,3*G+g] = -k4*NO[g]
        jac[2*G+g,3*G+g] = k4*NO[g]
        jac[3*G+g,3*G+g] = -k4*NO[g]

    return jac


    
# Hybrid integration using first-order splitting
# of the non-stiff part followed by the stiff part
def Hybrid(X, Dt):

    Nin = int(round(Dt/dt))

    for iter in range(Nin):
    
        # 4th order Runge-Kutta for the non-stiff equations

        z1 = KineticNonStiff(X)*dt
        z2 = KineticNonStiff(X + z1/2.)*dt
        z3 = KineticNonStiff(X + z2/2.)*dt
        z4 = KineticNonStiff(X + z3)*dt
        X += (z1+2.*z2+2.*z3+z4)/6.


        # Rosenbrock for the stiff equations (chemistry)

        # Convert Lorenz95 time units into min
        dt_target = dt*7200.
        dt_part = 0

        Z = Chemistry(X[G:], dt_target, dt_target, dt_part)

        ## Clipping (set negative concentrations to 0)
        X[G:] = np.maximum(0,Z)


# Hybrid integration using first-order splitting
# of the non-stiff part followed by the stiff part
def Chemistry(Z, dt_current, dt_target, dt_part):


    ##  Compute "dc/dt" => cel 
    cel = KineticStiff(Z)
        
    ## Compute the Jacobian matrix => jac  
    jac = Jacobian(Z)

    ##  Compute z1, first step of the 2sd-order Rosenbrock scheme
    ##  (1-gamma*dt*jac)z1 = cel(z)
    z1 = np.empty([Nsp*G])
    _jac = np.empty([Nsp,Nsp])
    _cel = np.empty([Nsp])
    for g in range(G):
        for i in range(Nsp):
            _cel[i] = cel[g+i*G]
            for j in range(Nsp):
                _jac[i,j] = jac[g+i*G,g+j*G]
        _z1 = np.linalg.solve(np.eye(Nsp)-gamma*dt_current*_jac, _cel)
        for i in range(Nsp):
            z1[g+i*G] = _z1[i]


    ## Compute z2, second step of the 2sd-order Rosenbrock scheme
    ## (1-gamma*Dt*jac)z2 = cel(c+Dt*z1)-2*z1
    cel = KineticStiff(Z+dt_current*z1)-2*z1

    z2 = np.empty([Nsp*G])
    for g in range(G):
        for i in range(Nsp):
            _cel[i] = cel[g+i*G]
            for j in range(Nsp):
                _jac[i,j] = jac[g+i*G,g+j*G]
        _z2 = np.linalg.solve(np.eye(Nsp)-gamma*dt_current*_jac, _cel)
        for i in range(Nsp):
            z2[g+i*G] = _z2[i]

    ## Sum up  the two contributions
    _Z = np.empty([Nsp*G])
    _Z = Z + 1.5*dt_current*z1 + 0.5*dt_current*z2;


    # Minimum time-step in min
    dt_min = 10

    if adaptive_option==True:

        # Adaptive criterion
        cdt_max = 0.

        ## Compute the condition for the apdative time step
        cdt = max(abs((z1+z2)/_Z))/2
        if cdt>cdt_max: 
            cdt_max = cdt
        cdt_max *= dt_current

        if cdt_max>1 and dt_current>dt_min:

            dt_new = dt_current*0.9/sqrt(cdt_max)
            dt_new = max(dt_new, dt_min);

            if dt_part+dt_new > dt_target:
                dt_new = dt_target - dt_part
                
            _Z = Chemistry(Z,dt_new,dt_target,dt_part)
            
        else:
            
            if (dt_part+dt_current<dt_target):
                
                dt_new = dt_current*0.9/sqrt(cdt_max)
                dt_new = max(dt_new,dt_min)

                if dt_part+dt_current+dt_new > dt_target:
                    dt_new = dt_target - (dt_part + dt_current)

                _Z = Chemistry(_Z,dt_new,dt_target,dt_part+dt_current)


    return _Z


####################
# Initialisation   #
####################

def NewState():

    x= np.empty([M])

    # Wind
    x[:G] = 3. + np.random.normal(0, 1., G)
    # ROC in ppb
    x[G:2*G] = 1.18  #*exp(np.random.normal(0, 1., G*Nsp))
    # NO in ppb
    x[2*G:3*G] = 10.
    # NO2 in ppb
    x[3*G:4*G] = 10.
    # O3 in ppb
    x[4*G:5*G] = 8
    # SGN in ppb
    x[5*G:6*G] = 1.

    return x


